/***************************************************/
/*                                                 */
/*      AMBER NVIDIA CUDA CPU IMPLEMENTATION       */
/*                 PMEMD VERSION                   */
/*                     2010                        */
/*                      by                         */
/*             Scott Le Grand (NVIDIA)             */
/*               Duncan Poole (NVIDIA)             */
/*                Ross Walker (SDSC)               */
/*                                                 */
/***************************************************/

#include "gputypes.h"
#include <math.h>

void clearCudaSimulation(cudaSimulation& sim)
{
    sim.atoms                   = 0;
    sim.grid                    = GRID;
    sim.gridBits                = GRIDBITS;
    sim.ntp                     = 0;
    sim.alpb                    = 0;
    sim.igb                     = 0;
    sim.scnb                    = (1.0 / 2.0);
    sim.scee                    = (1.0 / 1.2);
    sim.cut                     = 8.0;
    sim.cut2                    = sim.cut * sim.cut;
    sim.skinnb                  = 2.0f;
    sim.dielc                   = 1.0;
    sim.tol                     = 0.0001;
    sim.bUseVlimit              = false;
    sim.vlimit                  = 0.0;
    sim.massH                   = 1.008;
    sim.gb_alpha                = 0.0;
    sim.gb_beta                 = 0.0;
    sim.gb_gamma                = 0.0; 
    sim.gb_fs_max               = 2.0;
    sim.rgbmax                  = 25.0;
    sim.intdiel                 = 1.0;
    sim.extdiel                 = 78.5;
    sim.alpb_alpha              = 0.571412;
    sim.saltcon                 = 0.0;
    sim.surften                 = 0.005;
    sim.offset                  = 0.09;
    sim.gb_neckscale            = 0.361825;
    sim.gb_neckcut              = 2.8;
    sim.gb_neckoffset           = 1.0 - 0.09;
    sim.arad                    = 15.0;
    sim.a                       = 100.0;
    sim.b                       = 100.0;
    sim.c                       = 100.0;
    sim.alpha                   = 90.0;
    sim.beta                    = 90.0;
    sim.gamma                   = 90.0;
    sim.pi_vol_inv              = 0.0;
    sim.fac                     = 0.0;
    sim.fac2                    = 0.0;
    sim.is_orthog               = true;
    sim.paddedNumberOfAtoms     = 0;
    sim.pAtomX                  = NULL;
    sim.pAtomY                  = NULL;
    sim.pAtomXYSP               = NULL;
    sim.pAtomZ                  = NULL;
    sim.pAtomZSP                = NULL;
    sim.pAtomSigEps             = NULL;
    sim.pAtomRBorn              = NULL;
    sim.pAtomS                  = NULL;
    sim.pgb_alpha               = NULL;
    sim.pgb_beta                = NULL;
    sim.pgb_gamma               = NULL;
    sim.pAtomInvMass            = NULL;
    sim.pAtomMass               = NULL;
    sim.pAtomCharge             = NULL;
    sim.pAtomChargeSP           = NULL;
    sim.pReff                   = NULL;
    sim.pReffSP                 = NULL;
    sim.pPsi                    = NULL;
    sim.pTemp7                  = NULL;
    sim.pForce                  = NULL;
    sim.pForceX                 = NULL;
    sim.pForceY                 = NULL;
    sim.pForceZ                 = NULL;
    sim.pUllForce               = NULL;
    sim.pUllForceX              = NULL;
    sim.pUllForceY              = NULL;
    sim.pUllForceZ              = NULL;    
#ifdef MPI    
    sim.pInForce                = NULL;
    sim.pOutForce               = NULL;
    sim.pPMEForce               = NULL;  
    sim.pTemp7a                 = NULL;
    sim.pReffa                  = NULL;   
#endif
    sim.pVelX                   = NULL;
    sim.pVelY                   = NULL;
    sim.pVelZ                   = NULL;
    sim.pLVelX                  = NULL;
    sim.pLVelY                  = NULL;
    sim.pLVelZ                  = NULL;
    sim.randomSteps             = 0;
    sim.randomNumbers           = 0;
    sim.pRandomPos              = NULL;
    sim.pRandom                 = NULL;
    sim.pRandomX                = NULL;
    sim.pRandomY                = NULL;
    sim.pRandomZ                = NULL;
    sim.shakeConstraints        = 0;
    sim.shakeOffset             = 0;
    sim.fastShakeConstraints    = 0;
    sim.fastShakeOffset         = 0;
    sim.pShakeID                = NULL;
    sim.pShakeParm              = NULL;
    sim.pFastShakeID            = NULL;
    sim.bonds                   = 0;
    sim.bondAngles              = 0;
    sim.bondAngleOffset         = 0;
    sim.dihedrals               = 0;
    sim.dihedralOffset          = 0;   
    sim.nb14s                   = 0;
    sim.nb14Offset              = 0;
    sim.constraints             = 0;   
    sim.constraintOffset        = 0;
    sim.UBAngles                = 0;
    sim.UBAngleOffset           = 0;
    sim.impDihedrals            = 0;
    sim.impDihedralOffset       = 0;
    sim.cmaps                   = 0;
    sim.cmapOffset              = 0;
    sim.pBond                   = NULL;
    sim.pBondID                 = NULL;
    sim.pBondAngle              = NULL;
    sim.pBondAngleID1           = NULL;
    sim.pBondAngleID2           = NULL;
    sim.pDihedral1              = NULL;
    sim.pDihedral2              = NULL;
    sim.pDihedral3              = NULL;
    sim.pDihedralID1            = NULL;
    sim.pDihedralID2            = NULL;
    sim.pNb141                  = NULL;
    sim.pNb142                  = NULL;
    sim.pNb14ID                 = NULL;
    sim.pConstraint1            = NULL;
    sim.pConstraint2            = NULL;
    sim.pConstraintID           = NULL;
    sim.pUBAngle                = NULL;
    sim.pUBAngleID              = NULL;
    sim.pImpDihedral            = NULL;    
    sim.pImpDihedralID1         = NULL; 
    sim.pImpDihedralID2         = NULL;
    sim.pCmapID1                = NULL;
    sim.pCmapID2                = NULL;
    sim.pCmapID3                = NULL;
    sim.pCmapType               = NULL;
    sim.pCmapEnergy             = NULL;
    sim.EPs                     = 0;
    sim.EP11s                   = 0;
    sim.EP12s                   = 0;
    sim.EP21s                   = 0;
    sim.EP22s                   = 0;
    sim.EP11Offset              = 0;
    sim.EP12Offset              = 0;
    sim.EP21Offset              = 0;
    sim.EP22Offset              = 0;
    sim.pExtraPoint11Frame      = NULL;
    sim.pExtraPoint11Index      = NULL;
    sim.pExtraPoint11X          = NULL;
    sim.pExtraPoint11Y          = NULL;
    sim.pExtraPoint11Z          = NULL;
    sim.pExtraPoint12Frame      = NULL;
    sim.pExtraPoint12Index      = NULL;
    sim.pExtraPoint12X          = NULL;
    sim.pExtraPoint12Y          = NULL;
    sim.pExtraPoint12Z          = NULL;
    sim.pExtraPoint21Frame      = NULL;
    sim.pExtraPoint21Index      = NULL;
    sim.pExtraPoint21X1         = NULL;
    sim.pExtraPoint21Y1         = NULL;
    sim.pExtraPoint21Z1         = NULL;    
    sim.pExtraPoint21X2         = NULL;
    sim.pExtraPoint21Y2         = NULL;
    sim.pExtraPoint21Z2         = NULL;        
    sim.pExtraPoint22Frame      = NULL;
    sim.pExtraPoint22Index      = NULL;
    sim.pExtraPoint22X1         = NULL;
    sim.pExtraPoint22Y1         = NULL;
    sim.pExtraPoint22Z1         = NULL;    
    sim.pExtraPoint22X2         = NULL;
    sim.pExtraPoint22Y2         = NULL;
    sim.pExtraPoint22Z2         = NULL;     
    sim.pImageExtraPoint11Frame = NULL;
    sim.pImageExtraPoint11Index = NULL;
    sim.pImageExtraPoint12Frame = NULL;
    sim.pImageExtraPoint12Index = NULL;
    sim.pImageExtraPoint21Frame = NULL;
    sim.pImageExtraPoint21Index = NULL;    
    sim.pImageExtraPoint22Frame = NULL;
    sim.pImageExtraPoint22Index = NULL;         
    sim.pForceBuffer            = NULL;
    sim.pForceXBuffer           = NULL;
    sim.pForceYBuffer           = NULL;
    sim.pForceZBuffer           = NULL;        
    sim.pEnergyBuffer           = NULL;   
    sim.pReffBuffer             = NULL;
    sim.pSumdeijdaBuffer        = NULL;
    sim.pNeckMaxValPos          = NULL;
    sim.stride                  = 0;
    sim.stride4                 = 0;     
    sim.ew_coeff                = (PMEFloat)1.0;
    sim.ew_coeff2               = sim.ew_coeff * sim.ew_coeff;
    sim.negTwoEw_coeffRsqrtPI   = (PMEFloat)(-2.0 * sim.ew_coeff / sqrt(PI));
    sim.nfft1                   = 0;
    sim.nfft2                   = 0;
    sim.nfft3                   = 0;
    sim.fft_x_dim               = 0;
    sim.fft_y_dim               = 0;
    sim.fft_z_dim               = 0;
    sim.order                   = 4;
    sim.pPrefac1                = NULL;
    sim.pPrefac2                = NULL;
    sim.pPrefac3                = NULL;
    sim.pIFractX                = NULL;
    sim.pIFractY                = NULL;
    sim.pIFractZ                = NULL;
    sim.pThetaX                 = NULL;
    sim.pThetaY                 = NULL;
    sim.pThetaZ                 = NULL;
    sim.pDThetaX                = NULL;
    sim.pDThetaY                = NULL;
    sim.pDThetaZ                = NULL;
    sim.pXYZ_q                  = NULL;
    sim.plXYZ_q                 = NULL;
    sim.pXYZ_qt                 = NULL;
    
    sim.cells                   = 0;
    sim.cell                    = (PMEFloat)1.0;
    sim.xcell                   = (PMEDouble)1.0;
    sim.ycell                   = (PMEDouble)1.0;
    sim.zcell                   = (PMEDouble)1.0;
    sim.minCellX                = (PMEDouble)-2.0;
    sim.minCellY                = (PMEDouble)-2.0;
    sim.minCellZ                = (PMEDouble)-2.0;
    sim.maxCellX                = (PMEDouble)2.0;
    sim.maxCellY                = (PMEDouble)2.0;
    sim.maxCellZ                = (PMEDouble)2.0;    
    sim.oneOverXcells           = (PMEFloat)1.0;
    sim.oneOverYcells           = (PMEFloat)1.0;
    sim.oneOverZcells           = (PMEFloat)1.0;    
    sim.xcells                  = 0;
    sim.ycells                  = 0;
    sim.zcells                  = 0;
    sim.pNLbSkinTestFail        = NULL;
    sim.pNLCellHash             = NULL;
    sim.pImageIndex             = NULL;
    sim.pImageIndex2            = NULL;
    sim.pImageAtom              = NULL;
    sim.pImageAtom2             = NULL;
    sim.pImageAtomLookup        = NULL;
    sim.pImageHash              = NULL;
    sim.pAtomXYSaveSP           = NULL;
    sim.pAtomZSaveSP            = NULL;
    sim.pImageX                 = NULL;
    sim.pImageY                 = NULL;
    sim.pImageZ                 = NULL; 
    sim.pImageVelX              = NULL;
    sim.pImageVelY              = NULL;
    sim.pImageVelZ              = NULL;
    sim.pImageLVelX             = NULL;
    sim.pImageLVelY             = NULL;
    sim.pImageLVelZ             = NULL;
    sim.pImageMass              = NULL;
    sim.pImageCharge            = NULL;
    sim.pImageSigEps            = NULL; 
    sim.pImageOutputBuffers     = NULL;
    sim.pImageCellID            = NULL;    
    sim.pImageX2                = NULL;
    sim.pImageY2                = NULL;
    sim.pImageZ2                = NULL; 
    sim.pImageVelX2             = NULL;
    sim.pImageVelY2             = NULL;
    sim.pImageVelZ2             = NULL;
    sim.pImageLVelX2            = NULL;
    sim.pImageLVelY2            = NULL;
    sim.pImageLVelZ2            = NULL;
    sim.pImageMass2             = NULL;
    sim.pImageCharge2           = NULL;
    sim.pImageSigEps2           = NULL; 
    sim.pImageOutputBuffers2    = NULL;
    sim.pImageCellID2           = NULL;
    sim.pImageBondID            = NULL;
    sim.pImageBondAngleID1      = NULL;
    sim.pImageBondAngleID2      = NULL;
    sim.pImageDihedralID1       = NULL;
    sim.pImageDihedralID2       = NULL;
    sim.pImageNb14ID            = NULL;
    sim.pImageConstraintID      = NULL;
    sim.pImageUBAngleID         = NULL;
    sim.pImageImpDihedralID1    = NULL;
    sim.pImageImpDihedralID2    = NULL;
    sim.pImageCmapID1           = NULL;
    sim.pImageCmapID2           = NULL;
    sim.pImageCmapID3           = NULL;  
    sim.pImageSoluteAtomID      = NULL;
    sim.pImageSolventAtomID     = NULL;
    sim.pNLNonbondCellStartEnd  = NULL;
    sim.pNLExclusionStartCount  = NULL;
    sim.pNLExclusionList        = NULL; 
    sim.pNLAtomList             = NULL;  
    sim.pNLTotalOffset          = NULL; 
    sim.maxNonbondBuffers       = 0;
    sim.soluteMolecules         = 0;
    sim.soluteAtoms             = 0;
    sim.solventMolecules        = 0;
    sim.solventMoleculeStride   = 0;
    sim.pSoluteAtomMoleculeID   = NULL;
    sim.pSoluteAtomID           = NULL;
    sim.pSoluteAtomMass         = NULL;
    sim.pSoluteCOMX             = NULL;
    sim.pSoluteCOMY             = NULL;
    sim.pSoluteCOMZ             = NULL;  
    sim.pSoluteDeltaCOMX        = NULL;
    sim.pSoluteDeltaCOMY        = NULL;
    sim.pSoluteDeltaCOMZ        = NULL;  
    sim.pSoluteUllCOMX          = NULL;
    sim.pSoluteUllCOMY          = NULL;
    sim.pSoluteUllCOMZ          = NULL;      
    sim.pSoluteUllEKCOMX        = NULL;
    sim.pSoluteUllEKCOMY        = NULL;
    sim.pSoluteUllEKCOMZ        = NULL;          
    sim.pSoluteInvMass          = NULL;
    sim.pSolventAtomID          = NULL;
    sim.pSolventAtomMass1       = NULL;
    sim.pSolventAtomMass2       = NULL;
    sim.pSolventAtomMass3       = NULL;
    sim.pSolventAtomMass4       = NULL;
    sim.pSolventCOMX            = NULL;
    sim.pSolventCOMY            = NULL;
    sim.pSolventCOMZ            = NULL;
    sim.pSolventInvMass         = NULL;
    sim.pKineticEnergy          = NULL; 
    sim.constraintSoluteMolecules           = 0;
    sim.constraintSoluteMoleculeStride      = 0;
    sim.constraintSoluteAtoms               = 0;
    sim.constraintSolventMolecules          = 0;
    sim.pConstraintSoluteAtomMoleculeID     = NULL;
    sim.pConstraintSoluteAtomConstraintID   = NULL;
    sim.pConstraintSoluteAtomMass           = NULL;
    sim.pConstraintSoluteAtomX              = NULL;
    sim.pConstraintSoluteAtomY              = NULL;
    sim.pConstraintSoluteAtomZ              = NULL;
    sim.pConstraintSoluteCOMX               = NULL;
    sim.pConstraintSoluteCOMY               = NULL;
    sim.pConstraintSoluteCOMZ               = NULL;
    sim.pConstraintSoluteUllCOMX            = NULL;
    sim.pConstraintSoluteUllCOMY            = NULL;
    sim.pConstraintSoluteUllCOMZ            = NULL;     
    sim.pConstraintSoluteInvMass            = NULL;
    sim.pConstraintSolventAtoms             = NULL;
    sim.pConstraintSolventConstraint        = NULL;
    sim.pConstraintSolventAtomMass1         = NULL;
    sim.pConstraintSolventAtomX1            = NULL;
    sim.pConstraintSolventAtomY1            = NULL;
    sim.pConstraintSolventAtomZ1            = NULL;
    sim.pConstraintSolventAtomMass2         = NULL;
    sim.pConstraintSolventAtomX2            = NULL;
    sim.pConstraintSolventAtomY2            = NULL;
    sim.pConstraintSolventAtomZ2            = NULL;
    sim.pConstraintSolventAtomMass3         = NULL;
    sim.pConstraintSolventAtomX3            = NULL;
    sim.pConstraintSolventAtomY3            = NULL;
    sim.pConstraintSolventAtomZ3            = NULL;
    sim.pConstraintSolventAtomMass4         = NULL;
    sim.pConstraintSolventAtomX4            = NULL;
    sim.pConstraintSolventAtomY4            = NULL;
    sim.pConstraintSolventAtomZ4            = NULL;
    sim.pConstraintSolventCOMX              = NULL;
    sim.pConstraintSolventCOMY              = NULL;    
    sim.pConstraintSolventCOMZ              = NULL;
    sim.pConstraintSolventInvMass           = NULL;
    
    sim.rgbmax1i                = 1.0 / sim.rgbmax;
    sim.rgbmax2i                = sim.rgbmax1i * sim.rgbmax1i;
    sim.rgbmaxpsmax2            = (sim.rgbmax + sim.gb_fs_max) * (sim.rgbmax + sim.gb_fs_max);
    sim.alpb_beta               = sim.alpb_alpha * (sim.intdiel / sim.extdiel);
    sim.one_arad_beta           = sim.alpb_beta / sim.arad;
    if (sim.saltcon >= 0) 
    {
        sim.gb_kappa            = 0.73 * sqrt(0.10806 * sim.saltcon);
        if (sim.gb_kappa > 0.0)
            sim.gb_kappa_inv    = 1.0 / sim.gb_kappa;
    }
    else
    {
        sim.gb_kappa            = 0.0;
        sim.gb_kappa_inv        = 0.0;
    }
    if (sim.alpb == 0)
    {
        sim.intdiel_inv         = 1.0 / sim.intdiel;
        sim.extdiel_inv         = 1.0 / sim.extdiel;
    }
    else
    {
        sim.extdiel_inv         = 1.0 / (sim.extdiel * (1.0 + sim.alpb_beta));
        sim.intdiel_inv         = 1.0 / (sim.intdiel * (1.0 + sim.alpb_beta));
    }
    sim.invMassH                = 1.0 / sim.massH;
    
    sim.ucell[0][0]             = sim.a;
    sim.ucell[0][1]             = 0.0;
    sim.ucell[0][2]             = 0.0;
    sim.ucell[1][0]             = 0.0;
    sim.ucell[1][1]             = sim.b;
    sim.ucell[1][2]             = 0.0;
    sim.ucell[2][0]             = 0.0;
    sim.ucell[2][1]             = 0.0;
    sim.ucell[2][2]             = sim.c;
    sim.recip[0][0]             = 1.0 / sim.a;
    sim.recip[0][1]             = 0.0;
    sim.recip[0][2]             = 0.0;
    sim.recip[1][0]             = 0.0;
    sim.recip[1][1]             = 1.0 / sim.b;
    sim.recip[1][2]             = 0.0;
    sim.recip[2][0]             = 0.0;
    sim.recip[2][1]             = 0.0;
    sim.recip[2][2]             = 1.0 / sim.c;
    sim.pbc_box[0]              = sim.a;
    sim.pbc_box[1]              = sim.b;
    sim.pbc_box[2]              = sim.c;
    sim.reclng[0]               = 1.0 / sim.a;
    sim.reclng[1]               = 1.0 / sim.b;
    sim.reclng[2]               = 1.0 / sim.c;
    sim.uc_volume               = sim.a * sim.b * sim.c;
    sim.uc_sphere               = sqrt(sim.a * sim.a + sim.b * sim.b + sim.c * sim.c);
    sim.NLMaxTotalOffset        = 0;   
    sim.NLWorkUnits             = 0;         
}

_gpuContext::_gpuContext() :
bECCSupport(false),
totalCPUMemory(0),
totalGPUMemory(0),
#ifdef MPI
nGpus(1),
gpuID(0),
#endif
sm_version(SM_10),
ntb(1),
ntc(1),
ntf(1),
ntr(0),
forwardPlan(0),
backwardPlan(0),
pbAtom(NULL),
pbAtomXYSP(NULL),
pbAtomZSP(NULL),
pbAtomSigEps(NULL),
pbAtomCharge(NULL),
pbAtomChargeSP(NULL),
pbAtomRBorn(NULL),
pbAtomS(NULL),
pbAtomMass(NULL),
pbCenter(NULL),
pbForce(NULL),
pbUllForce(NULL),
#ifdef MPI
pbReffa(NULL),
pbTemp7a(NULL),
pbInForce(NULL),
pbOutForce(NULL),
pbPMEForce(NULL),
pPMEStart(NULL),
pPMELength(NULL),
pMinLocalCell(NULL),
pMaxLocalCell(NULL),
pMinLocalAtom(NULL),
pMaxLocalAtom(NULL),
pMinProcessedCell(NULL),
pMaxProcessedCell(NULL),
pMinProcessedAtom(NULL),
pMaxProcessedAtom(NULL),
pAllGathervRecvCountAoS(NULL),
pAllGathervRecvDisplAoS(NULL),
pAllGathervRecvCountSoA(NULL),
pAllGathervRecvDisplSoA(NULL),
pForceSendNode(NULL),
pForceSendMinCell(NULL),
pForceSendMaxCell(NULL),
pOutForceSendStart(NULL),
pForceSendStart(NULL),
pForceSendLength(NULL),     
pPMEForceData(NULL),
pSharedForceBuffer0(NULL),
pSharedForceBuffer1(NULL),
pSharedForceBuffer2(NULL),
pSharedPMEForceBuffer0(NULL),
pSharedPMEForceBuffer1(NULL),
pSharedPMEForceBuffer2(NULL),    
pSharedReffaBuffer(NULL),
pSharedTemp7aBuffer(NULL), 
#endif
pbVel(NULL),
pbLVel(NULL),
bLocalInteractions(true),
bCharmmInteractions(false),
pbBond(NULL),
pbBondID(NULL),
pbBondAngle(NULL),
pbBondAngleID1(NULL),
pbBondAngleID2(NULL),
pbDihedral1(NULL),
pbDihedral2(NULL),
pbDihedral3(NULL),
pbDihedralID1(NULL),
pbDihedralID2(NULL),
pbNb141(NULL),
pbNb142(NULL),
pbNb14ID(NULL),
pbConstraint1(NULL),
pbConstraint2(NULL),
pbConstraintID(NULL),
pbUBAngle(NULL),
pbUBAngleID(NULL),
pbImpDihedral(NULL),
pbImpDihedralID1(NULL), 
pbImpDihedralID2(NULL),
pbCmapID1(NULL),
pbCmapID2(NULL),
pbCmapID3(NULL),
pbCmapType(NULL),
pbCmapEnergy(NULL),
pbOutputBufferCounter(NULL),
pbReff(NULL),
pbReffSP(NULL),
pbPsi(NULL),
pbReffBuffer(NULL),
pbTemp7(NULL),
pbSumdeijdaBuffer(NULL),
pbForceBuffer(NULL),
pbEnergyBuffer(NULL),
pbKineticEnergyBuffer(NULL),
pbWorkUnit(NULL),
pbExclusion(NULL),
pbGBPosition(NULL),
pbNeckMaxValPos(NULL),
pbGBAlphaBetaGamma(NULL),
randomCounter(0),
pbRandomPos(NULL),
pbRandom(NULL),
pbShakeID(NULL),
pbShakeParm(NULL),
pbFastShakeID(NULL),
pbExtraPoint11Frame(NULL),
pbExtraPoint11Index(NULL),
pbExtraPoint11(NULL),
pbExtraPoint12Frame(NULL),
pbExtraPoint12Index(NULL),
pbExtraPoint12(NULL),
pbExtraPoint21Frame(NULL),
pbExtraPoint21Index(NULL),
pbExtraPoint21(NULL),
pbExtraPoint22Frame(NULL),
pbExtraPoint22Index(NULL),
pbExtraPoint22(NULL),
pbImageExtraPoint11Frame(NULL),
pbImageExtraPoint11Index(NULL),
pbImageExtraPoint12Frame(NULL),
pbImageExtraPoint12Index(NULL),
pbImageExtraPoint21Frame(NULL),
pbImageExtraPoint21Index(NULL),
pbImageExtraPoint22Frame(NULL),
pbImageExtraPoint22Index(NULL),
pbPrefac(NULL),
pbIFract(NULL),
pbTheta(NULL),
pbRSNLRecord(NULL),
pbXYZ_q(NULL),
pbXYZ_qt(NULL),
bNeighborList(false),
bNeedNewNeighborList(true),
bNewNeighborList(false),
bSmallBox(false),
bOddNLCells(false),
neighborListBits(32),
pbAtomXYSaveSP(NULL),
pbAtomZSaveSP(NULL),
pbImage(NULL),
pbImageIndex(NULL),
pbImageVel(NULL),
pbImageLVel(NULL),
pbImageMass(NULL),
pbImageCharge(NULL),
pbImageSigEps(NULL),
pbImageOutputBuffers(NULL),
pbImageCellID(NULL),
pbImageBondID(NULL),
pbImageBondAngleID1(NULL),      
pbImageBondAngleID2(NULL),      
pbImageDihedralID1(NULL),
pbImageDihedralID2(NULL),
pbImageNb14ID(NULL),  
pbImageConstraintID(NULL),
pbImageUBAngleID(NULL),
pbImageImpDihedralID1(NULL), 
pbImageImpDihedralID2(NULL),
pbImageCmapID1(NULL),
pbImageCmapID2(NULL),
pbImageCmapID3(NULL),
pbImageShakeID(NULL),
pbImageFastShakeID(NULL),
pbImageSolventAtomID(NULL),
pbImageSoluteAtomID(NULL),
pbNLExclusionList(NULL),
pbNLExclusionStartCount(NULL),
pbNLAtomList(NULL),
pbNLOffset(NULL),
pbNLTotalOffset(NULL),
pbNLPosition(NULL),
pbNLNonbondCellStartEnd(NULL),
pbNLChargeGridBufferOffset(NULL),
pbNLOddBufferOverlapFlag(NULL),
pbNLbSkinTestFail(NULL),
pbNLCellHash(NULL),
maxSoluteMolecules(0),
maxPSSoluteMolecules(0),
pbSoluteAtomID(NULL),
pbSoluteAtomMass(NULL),
pbSolute(NULL),
pbUllSolute(NULL),
pbSolventAtomID(NULL),
pbSolvent(NULL),
pbNTPData(NULL),
pbConstraintSoluteID(NULL),
pbConstraintSoluteAtom(NULL),
pbConstraintSolute(NULL),
pbConstraintUllSolute(NULL),
pbConstraintSolventAtoms(NULL),
pbConstraintSolventAtom(NULL),
pbConstraintSolventConstraint(NULL),
ee_plasma(0.0),
self_energy(0.0),
vdw_recip(0.0)
{
   clearCudaSimulation(sim);
}

_gpuContext::~_gpuContext()
{
    // Delete Atom data
    delete pbAtom;
    delete pbAtomXYSP;
    delete pbAtomZSP;    
    delete pbAtomSigEps;
    delete pbAtomRBorn;
    delete pbAtomS;
    delete pbAtomCharge;
    delete pbAtomChargeSP;
    delete pbAtomMass;
    delete pbReff;
    delete pbReffSP;
    delete pbPsi;
    delete pbTemp7;
    delete pbForce;
    delete pbUllForce;
#ifdef MPI
    delete[] pMinLocalCell;
    delete[] pMaxLocalCell;
    delete[] pMinProcessedCell;
    delete[] pMaxProcessedCell;
    delete[] pMinProcessedAtom;
    delete[] pMaxProcessedAtom;
    delete[] pAllGathervRecvCountAoS;
    delete[] pAllGathervRecvDisplAoS;
    delete[] pAllGathervRecvCountSoA;
    delete[] pAllGathervRecvDisplSoA;
    delete pbInForce;
    delete pbOutForce;
    delete pbPMEForce;
    delete pbTemp7a;
    delete pbReffa;
    delete[] pPMEStart;
    delete[] pPMELength;
    delete[] pForceSendNode;
    delete[] pForceSendMinCell;
    delete[] pForceSendMaxCell;
    delete[] pOutForceSendStart;  
    delete[] pForceSendStart;  
    delete[] pForceSendLength;      
#endif
    delete pbVel;
    delete pbLVel;
    delete pbCenter;
    delete pbOutputBufferCounter;
    
    // Delete PME stuff
    delete pbPrefac;
    delete pbIFract;
    delete pbTheta;
    delete pbRSNLRecord;
    delete pbXYZ_q;
    delete pbXYZ_qt;   
    cufftDestroy(forwardPlan);
    cufftDestroy(backwardPlan);
    
    // Delete neighbor list stuff
    delete pbAtomXYSaveSP;
    delete pbAtomZSaveSP;
    delete pbImage;
    delete pbImageIndex;
    delete pbImageVel;
    delete pbImageLVel;
    delete pbImageMass;
    delete pbImageCharge;
    delete pbImageSigEps;
    delete pbImageOutputBuffers;
    delete pbImageCellID;
    delete pbImageBondID;
    delete pbImageBondAngleID1;
    delete pbImageBondAngleID2;
    delete pbImageDihedralID1;
    delete pbImageDihedralID2;
    delete pbImageNb14ID;                
    delete pbImageConstraintID;
    delete pbImageUBAngleID;
    delete pbImageImpDihedralID1; 
    delete pbImageImpDihedralID2;
    delete pbImageCmapID1;
    delete pbImageCmapID2;
    delete pbImageCmapID3;
    delete pbImageShakeID;
    delete pbImageFastShakeID;
    delete pbImageSolventAtomID;
    delete pbImageSoluteAtomID;
    delete pbNLNonbondCellStartEnd;
    delete pbNLChargeGridBufferOffset;
    delete pbNLOddBufferOverlapFlag;
    delete pbNLExclusionList;
    delete pbNLExclusionStartCount;
    delete pbNLAtomList;
    delete pbNLOffset;
    delete pbNLTotalOffset;
    delete pbNLPosition;
    delete pbNLbSkinTestFail;
    delete pbNLCellHash;

    // Delete NTP stuff
    delete pbSoluteAtomID;
    delete pbSoluteAtomMass;
    delete pbSolute;
    delete pbUllSolute;
    delete pbSolventAtomID;
    delete pbSolvent;
    delete pbNTPData;
    delete pbConstraintSoluteID;
    delete pbConstraintSoluteAtom;
    delete pbConstraintSolute;
    delete pbConstraintUllSolute;
    delete pbConstraintSolventAtoms;
    delete pbConstraintSolventAtom;
    delete pbConstraintSolventConstraint;
    
    // Delete GB stuff
    delete pbWorkUnit;
    delete pbExclusion;
    delete pbGBPosition;
    delete pbNeckMaxValPos;
    delete pbGBAlphaBetaGamma;
    
    // Delete random number stuff
    delete pbRandomPos;
    delete pbRandom;

    // Delete bonded parameter data
    delete pbBond;
    delete pbBondID;
    delete pbBondAngle;
    delete pbBondAngleID1;
    delete pbBondAngleID2;
    delete pbDihedral1;
    delete pbDihedral2;
    delete pbDihedral3;
    delete pbDihedralID1;
    delete pbDihedralID2;
    delete pbNb141;
    delete pbNb142;
    delete pbNb14ID;
    delete pbConstraint1;
    delete pbConstraint2;
    delete pbConstraintID;
    delete pbUBAngle;
    delete pbUBAngleID;
    delete pbImpDihedral;
    delete pbImpDihedralID1; 
    delete pbImpDihedralID2;
    delete pbCmapID1;
    delete pbCmapID2;
    delete pbCmapID3;
    delete pbCmapType;
    delete pbCmapEnergy;
    
    // Delete Shake constraint data
    delete pbShakeID;
    delete pbShakeParm;
    delete pbFastShakeID;
    
    // Delete extra points data
    delete pbExtraPoint11Frame;
    delete pbExtraPoint11Index;
    delete pbExtraPoint11;
    delete pbExtraPoint12Frame;
    delete pbExtraPoint12Index;
    delete pbExtraPoint12;
    delete pbExtraPoint21Frame;
    delete pbExtraPoint21Index;
    delete pbExtraPoint21;  
    delete pbExtraPoint22Frame;
    delete pbExtraPoint22Index;
    delete pbExtraPoint22;   
    delete pbImageExtraPoint11Frame;
    delete pbImageExtraPoint11Index;
    delete pbImageExtraPoint12Frame;
    delete pbImageExtraPoint12Index;
    delete pbImageExtraPoint21Frame;
    delete pbImageExtraPoint21Index;
    delete pbImageExtraPoint22Frame;
    delete pbImageExtraPoint22Index;
    
    // Delete accumulation buffers
    delete pbReffBuffer;
    delete pbSumdeijdaBuffer;
    delete pbForceBuffer; 
    delete pbEnergyBuffer;
    delete pbKineticEnergyBuffer;
}
